"""Adds a new Address declaration
.. moduleauthor:: Louis Holbrook <dev@holbrook.no>
.. pgp:: 0826EDA1702D1E87C6E2875121D2E7BB88C2A746 

"""

# standard imports
import os
import logging

# external imports
import chainlib.eth.cli
from chainlib.chain import ChainSpec
from chainlib.eth.connection import EthHTTPConnection
from chainlib.eth.tx import receipt
from chainlib.eth.address import to_checksum_address
from hexathon import (
        add_0x,
        strip_0x,
        )
from chainlib.eth.cli.arg import (
        Arg,
        ArgFlag,
        process_args,
        )
from chainlib.eth.cli.config import (
        Config,
        process_config,
        )
from chainlib.eth.cli.log import process_log
from chainlib.eth.settings import process_settings
from chainlib.settings import ChainSettings


# local imports
from eth_address_declarator.declarator import AddressDeclarator

logging.basicConfig(level=logging.WARNING)
logg = logging.getLogger()


def process_config_local(config, arg, args, flags):
    hsh = strip_0x(config.get('_POSARG'))
    if len(hsh) != 64:
        raise ValueError('declaration hash must be 32 bytes')
    config.add(hsh, '_DECLARATION')
    return config


arg_flags = ArgFlag()
arg = Arg(arg_flags)
flags = arg_flags.STD_WRITE | arg_flags.WALLET | arg_flags.EXEC

argparser = chainlib.eth.cli.ArgumentParser()
argparser = process_args(argparser, arg, flags)
argparser.add_argument('declaration', type=str, help='SHA256 sum of endorsement data to add')
args = argparser.parse_args()

logg = process_log(args, logg)

config = Config()
config = process_config(config, arg, args, flags, positional_name='declaration')
config = process_config_local(config, arg, args, flags)
logg.debug('config loaded:\n{}'.format(config))

settings = ChainSettings()
settings = process_settings(settings, config)
logg.debug('settings loaded:\n{}'.format(settings))


def main():
    conn = settings.get('CONN')
    c = AddressDeclarator(
            settings.get('CHAIN_SPEC'),
            signer=settings.get('SIGNER'),
            gas_oracle=settings.get('FEE_ORACLE'),
            nonce_oracle=settings.get('NONCE_ORACLE'),
            )
    subject_address = to_checksum_address(settings.get('RECIPIENT'))
    contract_address = to_checksum_address(settings.get('EXEC'))

    (tx_hash_hex, o) = c.add_declaration(
            settings.get('EXEC'),
            settings.get('SENDER_ADDRESS'),
            settings.get('RECIPIENT'),
            add_0x(config.get('_DECLARATION')),
            )

    if settings.get('RPC_SEND'):
        conn.do(o)
        if config.true('_WAIT'):
            r = conn.wait(tx_hash_hex)
            if r['status'] == 0:
                sys.stderr.write('EVM revert while deploying contract. Wish I had more to tell you')
                sys.exit(1)

        print(tx_hash_hex)
    else:
        print(o)


if __name__ == '__main__':
    main()



# Author:	Louis Holbrook <dev@holbrook.no> 0826EDA1702D1E87C6E2875121D2E7BB88C2A746
# SPDX-License-Identifier:	GPL-3.0-or-later
# File-version: 1
# Description: Python interface to abi and bin files for faucet contracts

# standard imports
import logging
import json
import os
import hashlib

# external imports
from hexathon import (
        strip_0x,
        add_0x,
        )
from chainlib.eth.tx import (
        TxFormat,
        TxFactory,
        )
from chainlib.eth.contract import (
        ABIContractEncoder,
        ABIContractType,
        abi_decode_single,
        )
from chainlib.eth.constant import ZERO_ADDRESS

# local imports
from eth_address_declarator import Declarator

logg = logging.getLogger(__name__)

moddir = os.path.dirname(__file__)
datadir = os.path.join(moddir, 'data')


def to_declarator_key(declarator_address_hex, declaration_address_hex):
    h = hashlib.new('sha256')
    h.update(bytes.fromhex(strip_0x(declaration_address_hex)))
    h.update(bytes.fromhex(strip_0x(declarator_address_hex)))
    return h.digest()


class AddressDeclarator(Declarator):

    __abi = None
    __bytecode = None

    @staticmethod
    def abi():
        if AddressDeclarator.__abi == None:
            f = open(os.path.join(datadir, 'AddressDeclarator.json'), 'r')
            AddressDeclarator.__abi = json.load(f)
            f.close()
        return AddressDeclarator.__abi


    @staticmethod
    def bytecode():
        if AddressDeclarator.__bytecode == None:
            f = open(os.path.join(datadir, 'AddressDeclarator.bin'))
            AddressDeclarator.__bytecode = f.read()
            f.close()
        return AddressDeclarator.__bytecode


    @staticmethod
    def gas(code=None):
        return 2000000


    def constructor(self, sender_address, initial_description):
        code = AddressDeclarator.bytecode()
        enc = ABIContractEncoder()
        initial_description_hex = add_0x(initial_description)
        enc.bytes32(initial_description_hex)
        code += enc.get()
        tx = self.template(sender_address, None, use_nonce=True)
        tx = self.set_code(tx, code)
        return self.build(tx)



# standard imports
import os
import unittest
import json
import logging
import hashlib

# external imports
from chainlib.eth.unittest.ethtester import EthTesterCase
from chainlib.eth.contract import (
        ABIContractEncoder,
        ABIContractType,
        )
from chainlib.eth.nonce import RPCNonceOracle
from chainlib.eth.tx import receipt
from giftable_erc20_token import GiftableToken
from hexathon import (
    add_0x,
    strip_0x,
    )

# local imports
from eth_address_declarator.declarator import AddressDeclarator
from eth_address_declarator import Declarator
from eth_address_declarator.unittest import TestAddressDeclaratorBase

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()

testdir = os.path.dirname(__file__)

description = '0x{:<064s}'.format(b'foo'.hex())


class TestAddressDeclarator(TestAddressDeclaratorBase):

    def setUp(self):
        super(TestAddressDeclarator, self).setUp()
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)

       #c = GiftableToken(signer=self.signer, nonce_oracle=nonce_oracle, chain_id=self.chain_spec.chain_id())
        c = GiftableToken(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.constructor(self.accounts[0], 'BarToken', 'BAR', 6)
        self.rpc.do(o)

        o = receipt(tx_hash_hex)
        r = self.rpc.do(o)
        self.assertEqual(r['status'], 1)

        self.bar_token_address = r['contract_address']


    def test_basic(self):
       
        d = add_0x(os.urandom(32).hex())

        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[0], self.foo_token_address, d)
        self.rpc.do(o)

        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.foo_token_address, d)
        self.rpc.do(o)

        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[0], self.bar_token_address, d)
        self.rpc.do(o)

        o = c.declarator_count(self.address, self.foo_token_address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declarator_count(r), 2)

        o = c.declarator_count(self.address, self.bar_token_address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declarator_count(r), 1)


    def test_get_single_declaration(self):
        d = add_0x(os.urandom(32).hex())

        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.foo_token_address, d)
        self.rpc.do(o)
        o = receipt(tx_hash_hex)
        r = self.rpc.do(o)
        self.assertEqual(r['status'], 1)

        o = c.declaration(self.address, self.accounts[1], self.foo_token_address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        proofs = c.parse_declaration(r)
        self.assertEqual(proofs[0], strip_0x(d))


    def test_declaration(self):

        d = add_0x(os.urandom(32).hex())
        d_two = add_0x(os.urandom(32).hex())

        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.foo_token_address, d)
        self.rpc.do(o)

        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.foo_token_address, d_two)
        self.rpc.do(o)

        nonce_oracle = RPCNonceOracle(self.accounts[2], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[2], self.foo_token_address, d)
        self.rpc.do(o)

        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[2], self.bar_token_address, d)
        self.rpc.do(o)

        o = c.declaration(self.address, self.accounts[1], self.foo_token_address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        proofs = c.parse_declaration(r)
        self.assertEqual(proofs[0], strip_0x(d))
        self.assertEqual(proofs[1], strip_0x(d_two))


    def test_declarator_to_subject(self):
        d = add_0x(os.urandom(32).hex())

        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.foo_token_address, d)
        self.rpc.do(o)

        nonce_oracle = RPCNonceOracle(self.accounts[2], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[2], self.foo_token_address, d)
        self.rpc.do(o)

        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.bar_token_address, d)
        self.rpc.do(o)

        o = c.declaration_address_at(self.address, self.accounts[1], 0, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declaration_address_at(r), strip_0x(self.foo_token_address))

        o = c.declaration_address_at(self.address, self.accounts[2], 0, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declaration_address_at(r), strip_0x(self.foo_token_address))

        o = c.declaration_address_at(self.address, self.accounts[1], 1, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declaration_address_at(r), strip_0x(self.bar_token_address))


    def test_subject_to_declarator(self):
        d = '0x' + os.urandom(32).hex()

        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.foo_token_address, d)
        self.rpc.do(o)

        nonce_oracle = RPCNonceOracle(self.accounts[2], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[2], self.foo_token_address, d)
        self.rpc.do(o)

        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[1], self.bar_token_address, d)
        self.rpc.do(o)

        o = c.declarator_address_at(self.address, self.foo_token_address, 0, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declaration_address_at(r), strip_0x(self.accounts[1]))

        o = c.declarator_address_at(self.address, self.foo_token_address, 1, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declaration_address_at(r), strip_0x(self.accounts[2]))


    def test_three_first(self):
        d = []
        for i in range(3):
            d.append(add_0x(os.urandom(32).hex()))

        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        for proof in d:
            (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[0], self.foo_token_address, proof)
            self.rpc.do(o)

        o = c.declarator_count(self.address, self.foo_token_address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        self.assertEqual(c.parse_declarator_count(r), 1)

        o = c.declaration(self.address, self.accounts[0], self.foo_token_address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        proofs = c.parse_declaration(r)
        self.assertEqual(len(proofs), 3)

        for i in range(3):
            self.assertEqual(proofs[i], strip_0x(d[i]))


    def test_three_first_different(self):
        d = []
        a = []
        for i in range(3):
            d.append(add_0x(os.urandom(32).hex()))
            a.append(add_0x(os.urandom(20).hex()))

        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)
        c = Declarator(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        for i in range(3):
            (tx_hash_hex, o) = c.add_declaration(self.address, self.accounts[0], a[i], d[i])
            self.rpc.do(o)

        for i in range(3):
            o = c.declarator_count(self.address, a[i], sender_address=self.accounts[0])
            r = self.rpc.do(o)
            self.assertEqual(c.parse_declarator_count(r), 1)

            o = c.declaration(self.address, self.accounts[0], a[i], sender_address=self.accounts[0])
            r = self.rpc.do(o)
            proofs = c.parse_declaration(r)
            self.assertEqual(len(proofs), 1)
            self.assertEqual(proofs[0], strip_0x(d[i]))


if __name__ == '__main__':
    unittest.main()
